@@ -1,5 +1,6 @@
-/* install-mbr for installing a master boot record
-   Copyright (C) 1999 Neil Turton
+/* install-mbr for installing a master boot record */
+#define COPYRIGHT "Copyright (C) 1999-2001 Neil Turton."
+/* See file AUTHORS for a list of contributors.
 
    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
@@ -27,9 +28,12 @@ See file COPYING for details */
 #include <fcntl.h>
 #include <errno.h>
 #include <assert.h>
-typedef unsigned char uchar;
+#include <time.h>
 #include "mbr.h"
 
+typedef u_int8_t  ui8_t;
+typedef u_int16_t ui16_t;
+
 #define BUFFER_SIZE 512
 #define CODE_SIZE (512-66)
 #define PTN_TABLE_OFFSET CODE_SIZE
@@ -40,44 +44,71 @@ typedef unsigned char uchar;
 
 const char *prog_name="install-mbr";
 
+/* There are 3 routines (and their helpers) which need to know ALL the
+ * members of this structure.  These are: modify_params fetch_params
+ * and merge_changes.*/
 struct change_params {
-  uchar flags;
-#define CPFLAG_DRIVE         1
-#define CPFLAG_DEFAULT       2
-#define CPFLAG_TIMEOUT_VALID 4
-  /* The following bit is used to solve conflicts between "-t <timeout>"
-   * and "-i a".  It is set by "-t <timeout>" and cleared by "-i a".
-   * If it is set, "-i a" is ignored. */
-#define CPFLAG_TIMEOUT       8
-#define CPFLAG_ENABLED      16
-#define CPFLAG_INTERRUPT    32
-  uchar drive;
-  uchar deflt;
-  uchar timeout;
-  uchar enabled, enabled_mask;
-#define CPE_1      ((uchar)1)
-#define CPE_2      ((uchar)2)
-#define CPE_3      ((uchar)4)
-#define CPE_4      ((uchar)8)
-#define CPE_F      ((uchar)16)
-#define CPE_A      ((uchar)128)
-  uchar interrupt, interrupt_mask;
+  /* Version numbers can't be changed... */
+  ui8_t  compat_version;
+  ui8_t  version;
+  /* This now describes which fields are valid */
+  ui8_t flags;
+#define CPFLAG_DRIVE              1
+#define CPFLAG_DEFAULT            2
+#define CPFLAG_TIMEOUT_VALID      4
+  /* The following bit is used to resolve conflicts between
+   * "-t <timeout>" and "-i a".  It is set by "-t <timeout>" and
+   * cleared by "-i a".  If it is set, "-i a" is ignored.  This allows
+   * "-t <timeout>" followed by "-i a-a" to work. */
+#define CPFLAG_TIMEOUT_OVERRIDES  8
+#define CPFLAG_ENABLED           16
+#define CPFLAG_INTERRUPT         32
+#define CPFLAG_Y2KBUG            64 /* Selects default code. */
+#define CPFLAG_Y2KDATE          128 /* Requires CPFLAG_Y2KBUG */
+
+  /* Requires CPFLAG_DRIVE */
+  ui8_t  drive;
+
+  /* Requires CPFLAG_DEFAULT */
+  ui8_t  deflt;
+
+  /* Requires CPFLAG_TIMEOUT_VALID */
+  ui16_t timeout;
+
+  /* Must be initialised.  Can be ignored unless CPFLAG_ENABLED */
+  ui8_t  enabled, enabled_mask;
+#define CPE_1      ((ui8_t)1)
+#define CPE_2      ((ui8_t)2)
+#define CPE_3      ((ui8_t)4)
+#define CPE_4      ((ui8_t)8)
+#define CPE_F      ((ui8_t)16)
+#define CPE_A      ((ui8_t)128)
+
+  /* Must be initialised.  Can be ignored unless CPFLAG_INTERRUPT */
+  ui8_t  interrupt, interrupt_mask;
 #define CPI_SHIFT  1
 #define CPI_KEY   2
 #define CPI_ALWAYS 4
 #define CPI_MASK   (CPI_SHIFT|CPI_KEY|CPI_ALWAYS)
+
+  /* Requires CPFLAG_Y2KBUG */
+  ui8_t  y2k_fix;
+
+  /* Requires CPFLAG_Y2KDATE */
+  ui8_t  y2k_month;
+  ui16_t y2k_year;
 };
 
-struct desc {
-  uchar *data;
+struct file_desc {
+  ui8_t *data;
   char *path;
   int flags;
-#define DESC_FLAG_FREE_DATA 1 /* Should we free the data? */
-#define DESC_FLAG_SPECIFIED 2
+#define FILE_DESC_FLAG_FREE_DATA 1 /* Should we free the data? */
+#define FILE_DESC_FLAG_SPECIFIED 2
 };
 
 struct data {
-  struct desc install, params, target, table;
+  struct file_desc install, params, target, table;
   unsigned int offset;
   int flags;
 #define DFLAG_KEEP     1
@@ -91,12 +122,14 @@ struct data {
 
 /* Version specific routines *********************************************/
 
-static struct mbr_params *
-find_params(const struct desc *dsc, int quiet)
+/* Locating parameters */
+
+static struct mbr_params_v1 *
+find_params_v1(const struct file_desc *dsc, int quiet)
 {
-  uchar *mbr=dsc->data;
+  ui8_t *mbr=dsc->data;
   uint ptr;
-  struct mbr_params *params;
+  struct mbr_params_v1 *params;
   uint ver_compat;
 
 #if 0
@@ -104,8 +137,8 @@ find_params(const struct desc *dsc, int quiet)
 #endif
 
   ptr=mbr[SIG_PTR_OFFSET]+256*mbr[SIG_PTR_OFFSET+1];
-  if(ptr>SIG_PTR_OFFSET-sizeof(struct mbr_params) ||
-     memcmp(mbr+ptr, MP_SIG, sizeof(params->sig))!=0)
+  if(ptr>SIG_PTR_OFFSET-(6+4+4) ||
+     memcmp(mbr+ptr, MP_V1_SIG, sizeof(params->sig))!=0)
   {
     /* Due to a silly little mistake, version 1.0.0 got released with
        the signature pointer uninitialised.  This means that everyone
@@ -113,53 +146,94 @@ find_params(const struct desc *dsc, int quiet)
        at 0x13b */
     ptr=0x13b;
     /* Parameters not found. */
-    if(memcmp(mbr+ptr, MP_SIG, sizeof(MP_SIG)))
-    {
-      if(quiet)
+    if(memcmp(mbr+ptr, MP_V1_SIG, sizeof(MP_V1_SIG)))
 	return NULL;
-
-      fprintf(stderr, "%s:%s: Failed to find MBR signature.\n",
-	      prog_name, dsc->path);
-      exit(1);
-    }
   }
-  params=(struct mbr_params *)(mbr+ptr);
-  ver_compat=params->ver_compat[0]+256*params->ver_compat[1];
-  if(ver_compat > MP_VERSION)
+  params=(struct mbr_params_v1 *)(mbr+ptr);
+  ver_compat=params->ver_compat;
+  if(ver_compat > MP_V1_VERSION)
   {
-    uint version=params->version[0]+256*params->version[1];
+    uint version=params->version;
 
-    if(quiet)
-      return NULL;
-
-    fprintf(stderr, "%s:%s: Cannot handle MBR version %x "
-	    "(backwards compatible to %x)\n",
+    /* FIXME: We should be able to override this error message. */
+    fprintf(stderr, "%s:%s: Cannot handle MBR version %d "
+	    "(backwards compatible to %d)\n",
 	    prog_name, dsc->path, version, ver_compat);
     exit(1);
   }
   return params;
 }
 
-static void
-modify_params(struct desc *dsc, struct change_params *cpp)
+static struct mbr_params_v2 *
+find_params_v2(const struct file_desc *dsc, int quiet)
 {
-  struct mbr_params *params=find_params(dsc, 0);
-  assert(MP_VERSION==0);
+  /* The parameters end just before the partition table, at 512-66. */
+  static const int offset = (512-66-sizeof(struct mbr_params_v2));
+  static const char sig[] = MP_V2_SIG;
+  struct mbr_params_v2 *params = (void*)(dsc->data + offset);
 
+  /* Check the signature. */
+  if (memcmp(params->sig, sig, sizeof(sig)))
+    return NULL;
+
+  /* Check the version. */
+  if (params->version < MP_V2_VERSION_MIN ||
+      params->ver_compat < MP_V2_VERSION_MIN ||
+      params->ver_compat > MP_V2_VERSION)
+  {
+    /* FIXME: We should be able to override this error message. */
+    fprintf(stderr, "%s:%s: Cannot handle MBR version %d "
+	    "(backwards compatible to %d)\n",
+	    prog_name, dsc->path, params->version, params->ver_compat);
+    exit(1);
+  }
+
+  return params;
+}
+
+/*************************************************************************/
+
+/* Modifying parameters.  First, we have some individual routines to
+   set specific fields. */
+
+/* This has no power to change anything.  It just makes sure that the
+   request is compatible. */
+static void
+check_variant(ui8_t variant, const struct change_params *cpp)
+{
+  if (cpp->flags & CPFLAG_Y2KBUG)
+  {
+    /* Make sure the settings match. */
+    if (cpp->y2k_fix != ((variant & MP_VARIANT_Y2K) ? 1 : 0))
+    {
+      fprintf(stderr, "%s: Selected code does not have requested "
+	      "Y2K bug fix setting.\n", prog_name);
+      exit(1);
+    }
+  }
+}
+
+static void
+modify_drive(ui8_t *drivep, const struct change_params *cpp)
+{
   if(cpp->flags & CPFLAG_DRIVE)
-    params->drive=cpp->drive;
+    *drivep=cpp->drive;
+}
 
+static void
+modify_default(ui8_t *defaultp, const struct change_params *cpp)
+{
   if(cpp->flags & CPFLAG_DEFAULT)
   {
-    params->deflt&=~MP_DEFLT_BITS;
+    *defaultp&=~MP_DEFLT_BITS;
     switch(cpp->deflt)
     {
-    case '1': params->deflt|=0; break;
-    case '2': params->deflt|=1; break;
-    case '3': params->deflt|=2; break;
-    case '4': params->deflt|=3; break;
-    case 'F': case 'f': params->deflt|=4; break;
-    case 'D': case 'd': params->deflt|=7; break;
+    case '1':           *defaultp|=0; break;
+    case '2':           *defaultp|=1; break;
+    case '3':           *defaultp|=2; break;
+    case '4':           *defaultp|=3; break;
+    case 'F': case 'f': *defaultp|=4; break;
+    case 'D': case 'd': *defaultp|=7; break;
 
     default:
       fprintf(stderr, "%s: Invalid default partition: %c\n",
@@ -167,6 +241,22 @@ modify_params(struct desc *dsc, struct change_params *cpp)
       exit(1);
     };
   }
+
+  /* The interrupt flags are stored here, too. */
+  if(cpp->interrupt & CPI_SHIFT)
+    *defaultp|=MP_DEFLT_ISHIFT;
+  else if((cpp->interrupt_mask & CPI_SHIFT)==0)
+    *defaultp&=~MP_DEFLT_ISHIFT;
+
+  if(cpp->interrupt & CPI_KEY)
+    *defaultp|=MP_DEFLT_IKEY;
+  else if((cpp->interrupt_mask & CPI_KEY)==0)
+    *defaultp&=~MP_DEFLT_IKEY;
+}
+
+static ui16_t
+modify_timeout_sub(ui16_t timeout, const struct change_params *cpp)
+{
   /* The interaction of the timeout with the interrupt always bit is a
      little subtle.
      CPIA_MASK CPIA TIMEOUT TIMEOUT_VAL comment
@@ -178,48 +268,174 @@ modify_params(struct desc *dsc, struct change_params *cpp)
   */
 
   if((cpp->interrupt & CPI_ALWAYS) != 0 /* If setting CPI_ALWAYS */
-     && (cpp->flags & CPFLAG_TIMEOUT) == 0) /* and not overridden */
-    params->delay=0xff; /* Set to always interrupt */
+     && (cpp->flags & CPFLAG_TIMEOUT_OVERRIDES) == 0) /* and not overridden */
+    timeout=0xffff; /* Set to always interrupt */
   else if(cpp->flags & CPFLAG_TIMEOUT_VALID)
-    params->delay=cpp->timeout;
+    timeout=cpp->timeout;
   /* Timeout is not valid and not setting CPI_ALWAYS, so the only
      action we can be taking is clearing CPI_ALWAYS */
   else if((cpp->interrupt_mask & CPI_ALWAYS)==0)
-    params->delay=DEFAULT_TIMEOUT;
+    timeout=DEFAULT_TIMEOUT;
 
-  if(cpp->interrupt & CPI_SHIFT)
-    params->deflt|=MP_DEFLT_ISHIFT;
-  else if((cpp->interrupt_mask & CPI_SHIFT)==0)
-    params->deflt&=~MP_DEFLT_ISHIFT;
+  return timeout;
+}
 
-  if(cpp->interrupt & CPI_KEY)
-    params->deflt|=MP_DEFLT_IKEY;
-  else if((cpp->interrupt_mask & CPI_KEY)==0)
-    params->deflt&=~MP_DEFLT_IKEY;
+static void
+modify_timeout_8(ui8_t *timeout, const struct change_params *cpp)
+{
+  ui16_t new_timeout = modify_timeout_sub(*timeout, cpp);
+  /* Make sure the value is valid for the old MBR.  Valid values are
+     0..0xfe and 0xffff. */
+  if (new_timeout > 0xfe && new_timeout != 0xffff)
+  {
+    fprintf(stderr, "%s: Timeout (%d) too large for legacy MBR.  "
+	    "Please upgrade.\n", prog_name, new_timeout);
+    exit(1);
+  }
 
+  *timeout = new_timeout;
+}
+
+static void
+modify_timeout_16(ui8_t timeout[2], const struct change_params *cpp)
+{
+  ui16_t delay = timeout[0] + 256*(ui16_t)timeout[1];
+  delay = modify_timeout_sub(delay, cpp);
+  timeout[0] = (delay>>0) & 0xff;
+  timeout[1] = (delay>>8) & 0xff;
+}
+
+static void
+modify_flags(ui8_t *flagsp, const struct change_params *cpp)
+{
   assert(MP_FLAG_EN1 == CPE_1);
   assert(MP_FLAG_EN2 == CPE_2);
   assert(MP_FLAG_EN3 == CPE_3);
   assert(MP_FLAG_EN4 == CPE_4);
   assert(MP_FLAG_ENF == CPE_F);
   assert(MP_FLAG_ENA == CPE_A);
-  params->flags=(params->flags & cpp->enabled_mask)|cpp->enabled;
+  *flagsp=(*flagsp & cpp->enabled_mask)|cpp->enabled;
 }
 
-/* Fetch the parameters from src to cp.  If quiet is set, return true
-   on failure, instead of calling exit(1). */
-static int
-fetch_params(struct change_params *cpp, const struct desc *src, int quiet)
+static void
+modify_date(ui8_t *date_ptr, const struct change_params *cpp)
 {
-  const struct mbr_params *params=find_params(src, quiet);
-  assert(MP_VERSION==0);
-  if(params==NULL)
-    return 1;
-  
-  cpp->flags=CPFLAG_DRIVE|CPFLAG_DEFAULT|CPFLAG_ENABLED|CPFLAG_INTERRUPT;
+  ui8_t month = cpp->y2k_month;
+  ui16_t year = cpp->y2k_year;
+  ui8_t year_hi, year_lo;
+  if (! (cpp->flags & CPFLAG_Y2KDATE))
+    return;
 
-  cpp->drive=params->drive;
-  switch(params->deflt & MP_DEFLT_BITS)
+  assert (cpp->flags & CPFLAG_Y2KBUG);
+  assert (cpp->y2k_fix);
+
+  /* These are stored in BCD (yawn). */
+  date_ptr[2]=(month/10<<4)|(month%10);
+  year_hi = year/100;
+  year_lo = year%100;
+  date_ptr[0]=(year_lo/10<<4)|(year_lo%10);
+  date_ptr[1]=(year_hi/10<<4)|(year_hi%10);
+}
+
+/* This is simply a case of applying the above routines. */
+static int
+modify_params_v1(struct file_desc *dsc, const struct change_params *cpp)
+{
+  struct mbr_params_v1 *params=find_params_v1(dsc, 0);
+  assert(MP_V1_VERSION==2);
+
+  if (params == NULL)
+    return 0;
+
+  check_variant(params->variant, cpp);
+
+  /* This section of structure changes format. */
+  if (params->version < MP_V1_FMTB_VER)
+  {
+    modify_drive(&(params->u.fmt_a.drive), cpp);
+    modify_timeout_8(&(params->u.fmt_a.delay), cpp);
+  }
+  else
+  {
+    modify_drive(&(params->u.fmt_b.drive), cpp);
+    modify_timeout_16(params->u.fmt_b.delay, cpp);
+  }
+  modify_default(&(params->deflt), cpp);
+  modify_flags(&(params->flags), cpp);
+  modify_date(dsc->data+SIG_PTR_OFFSET-3, cpp);
+
+  return 1;
+}
+
+static int
+modify_params_v2(struct file_desc *dsc, const struct change_params *cpp)
+{
+  struct mbr_params_v2 *params=find_params_v2(dsc, 0);
+  assert(MP_V2_VERSION==2);
+
+  if (params == NULL)
+    return 0;
+
+  check_variant(params->variant, cpp);
+  modify_drive(&(params->drive), cpp);
+  modify_default(&(params->deflt), cpp);
+  modify_timeout_16(params->delay, cpp);
+  modify_flags(&(params->flags), cpp);
+  modify_date(params->date, cpp);
+
+  return 1;
+}
+
+static void
+modify_params(struct file_desc *dsc, const struct change_params *cpp)
+{
+  if (modify_params_v1(dsc, cpp) ||
+      modify_params_v2(dsc, cpp))
+    return;
+
+  /* We can't continue. */
+  fprintf(stderr, "%s:%s: Failed to find MBR signature.\n",
+	  prog_name, dsc->path);
+  exit(1);
+}
+
+/* Fetching parameters.  First, we have some helper routines. */
+
+/* This must be called before the others. */
+static void
+handle_version(struct change_params *cpp, ui8_t compat, ui8_t ver)
+{
+  cpp->compat_version = compat;
+  cpp->version = ver;
+  cpp->flags = CPFLAG_INTERRUPT;
+  cpp->interrupt=0;
+  cpp->interrupt_mask=~0;
+  cpp->enabled=0;
+  cpp->enabled_mask=~0;
+}
+
+static void
+handle_variant(struct change_params *cpp, ui8_t variant)
+{
+  if (cpp->version >= MP_VERSION_Y2K)
+  {
+    cpp->flags |= CPFLAG_Y2KBUG;
+    cpp->y2k_fix = (variant & MP_VARIANT_Y2K) ? 1 : 0;
+  }
+}
+
+static void
+handle_drive(struct change_params *cpp, ui8_t drive)
+{
+  cpp->flags|=CPFLAG_DRIVE;
+  cpp->drive=drive;
+}
+
+static void
+handle_default(struct change_params *cpp, ui8_t deflt)
+{
+  cpp->flags|=CPFLAG_DEFAULT;
+  switch(deflt & MP_DEFLT_BITS)
   {
   case 0: cpp->deflt='1'; break;
   case 1: cpp->deflt='2'; break;
@@ -229,104 +445,226 @@ fetch_params(struct change_params *cpp, const struct desc *src, int quiet)
   case 7: cpp->deflt='D'; break;
   default:
       fprintf(stderr, "%s: Invalid default partition found: %d\n",
-	      prog_name, params->deflt);
+	      prog_name, deflt);
       exit(1);
   };
 
-  if(params->delay == 0xff)
+  cpp->interrupt_mask&=~(CPI_SHIFT|CPI_KEY);
+  if(deflt & MP_DEFLT_ISHIFT)
+    cpp->interrupt|=CPI_SHIFT;
+  if(deflt & MP_DEFLT_IKEY)
+    cpp->interrupt|=CPI_KEY;
+}
+
+static void
+handle_timeout_8(struct change_params *cpp, ui8_t delay)
+{
+  cpp->interrupt_mask&=~CPI_ALWAYS;
+  if(delay == 0xff)
   {
-    cpp->interrupt=CPI_ALWAYS;
+    cpp->interrupt|=CPI_ALWAYS;
   }
   else
   {
-    cpp->timeout=params->delay;
-    cpp->flags|=CPFLAG_TIMEOUT_VALID|CPFLAG_TIMEOUT;
-    cpp->interrupt=0;
+    cpp->timeout=delay;
+    cpp->flags|=CPFLAG_TIMEOUT_VALID|CPFLAG_TIMEOUT_OVERRIDES;
   }
-  cpp->interrupt_mask=~(CPI_ALWAYS|CPI_SHIFT|CPI_KEY);
-  if(params->deflt & MP_DEFLT_ISHIFT)
-    cpp->interrupt|=CPI_SHIFT;
-  if(params->deflt & MP_DEFLT_IKEY)
-    cpp->interrupt|=CPI_KEY;
+}
 
+static void
+handle_timeout_16(struct change_params *cpp, const ui8_t delay_p[2])
+{
+  u_int16_t delay = delay_p[0] + 256*(ui16_t)delay_p[1];
+  cpp->interrupt_mask&=~CPI_ALWAYS;
+  if(delay == 0xffff)
+  {
+    cpp->interrupt|=CPI_ALWAYS;
+  }
+  else
+  {
+    cpp->timeout=delay;
+    cpp->flags|=CPFLAG_TIMEOUT_VALID|CPFLAG_TIMEOUT_OVERRIDES;
+  }
+}
+
+static void
+handle_enable(struct change_params *cpp, ui8_t flags)
+{
   assert(MP_FLAG_EN1 == CPE_1);
   assert(MP_FLAG_EN2 == CPE_2);
   assert(MP_FLAG_EN3 == CPE_3);
   assert(MP_FLAG_EN4 == CPE_4);
   assert(MP_FLAG_ENF == CPE_F);
   assert(MP_FLAG_ENA == CPE_A);
-  cpp->enabled=params->flags;
-  cpp->enabled_mask=(uchar)~(uchar)(CPE_1|CPE_2|CPE_3|CPE_4|CPE_F|CPE_A);
-  return 0;
+  cpp->enabled=flags;
+  cpp->enabled_mask=(ui8_t)~(ui8_t)(CPE_1|CPE_2|CPE_3|CPE_4|CPE_F|CPE_A);
+  cpp->flags|=CPFLAG_ENABLED;
 }
 
+/* Fetch the parameters from src to cp.  If quiet is set, return false
+   on failure, instead of calling exit(1). */
+static int
+fetch_params_v1(struct change_params *cpp, const struct file_desc *src)
+{
+  const struct mbr_params_v1 *params=find_params_v1(src, 1);
+  assert(MP_V1_VERSION==2);
+  if(params==NULL)
+    return 0;
+  
+  handle_version(cpp, params->ver_compat, params->version);
+  handle_variant(cpp, params->variant);
+  /* This section of structure changes format. */
+  if (params->version < MP_V1_FMTB_VER)
+  {
+    handle_drive(cpp, params->u.fmt_a.drive);
+    handle_timeout_8(cpp, params->u.fmt_a.delay);
+  }
+  else
+  {
+    handle_drive(cpp, params->u.fmt_b.drive);
+    handle_timeout_16(cpp, params->u.fmt_b.delay);
+  }
+  handle_default(cpp, params->deflt);
+  handle_enable(cpp, params->flags);
+  return 1;
+}
+
+static int
+fetch_params_v2(struct change_params *cpp, const struct file_desc *src)
+{
+  const struct mbr_params_v2 *params=find_params_v2(src, 1);
+  assert(MP_V2_VERSION==2);
+  if(params==NULL)
+    return 0;
+  
+  handle_version(cpp, params->ver_compat, params->version);
+  handle_variant(cpp, params->variant);
+  handle_drive(cpp, params->drive);
+  handle_default(cpp, params->deflt);
+  handle_timeout_16(cpp, params->delay);
+  handle_enable(cpp, params->flags);
+  return 1;
+}
+
+static int
+fetch_params(struct change_params *cpp, const struct file_desc *src, int quiet)
+{
+  if (fetch_params_v1(cpp, src) ||
+      fetch_params_v2(cpp, src))
+    return 1;
+
+  /* We didn't find anything. */
+  if (quiet)
+    return 0;
+
+  /* We can't continue. */
+  fprintf(stderr, "%s:%s: Failed to find MBR signature.\n",
+	  prog_name, src->path);
+  exit(1);
+}
+
+
 /* Data handling *********************************************************/
 
+/* This copies values from src to dest.  Any value which is specified
+   in either original structure will be specified in the result.
+   Values in src override values in dest. */
 static void
-show_mbr(struct desc *dsc)
+merge_changes(struct change_params *dest, const struct change_params *src)
 {
-  struct change_params cp;
-  fetch_params(&cp, dsc, 0);
-  if(cp.flags & CPFLAG_DRIVE)
-    printf("Drive: 0x%x\n", cp.drive);
-  if(cp.flags & CPFLAG_DEFAULT)
-    printf("Default: %c\n", cp.deflt);
-  if(cp.flags & CPFLAG_TIMEOUT)
-    printf("Timeout: %d/18 seconds\n", cp.timeout);
-  if(cp.flags & CPFLAG_ENABLED)
+  int src_flags = src->flags;
+
+  /* We ignore the version numbers.  The dest should be right. */
+
+  /* Merge the flags.  This is usually the right thing to do, but see
+     below. */
+  dest->flags |= src_flags;
+
+  if (src_flags & CPFLAG_DRIVE)
+    dest->drive = src->drive;
+
+  if (src_flags & CPFLAG_DEFAULT)
+    dest->deflt = src->deflt;
+  
+  if (src_flags & CPFLAG_TIMEOUT_VALID)
+    dest->timeout = src->timeout;
+
+  /* If the source timeout doesn't override and interrupt has been
+     specified, make it so in the result. */
+  if ((src->interrupt & CPI_ALWAYS) != 0
+      && (src_flags & CPFLAG_TIMEOUT_OVERRIDES) == 0)
+    dest->flags &= ~CPFLAG_TIMEOUT_OVERRIDES;
+
+  dest->enabled &= src->enabled_mask;
+  dest->enabled_mask &= src->enabled_mask;
+  dest->enabled |= src->enabled;
+
+  dest->interrupt &= src->interrupt_mask;
+  dest->interrupt_mask &= src->interrupt_mask;
+  dest->interrupt |= src->interrupt;
+
+  if (src_flags & CPFLAG_Y2KBUG)
+    dest->y2k_fix = src->y2k_fix;
+
+  if (src_flags & CPFLAG_Y2KDATE)
+  {
+    dest->y2k_month = src->y2k_month;
+    dest->y2k_year  = src->y2k_year;
+  }
+}
+
+static show_params(const struct change_params *cp)
+{
+  printf("Version: %d\n", cp->version);
+  printf("Compatible: %d\n", cp->compat_version);
+  if(cp->flags & CPFLAG_Y2KBUG)
+    printf("Y2K-Fix: %s\n", cp->y2k_fix ? "Enabled" : "Disabled");
+  if(cp->flags & CPFLAG_DRIVE)
+    printf("Drive: 0x%x\n", cp->drive);
+  if(cp->flags & CPFLAG_DEFAULT)
+    printf("Default: %c\n", cp->deflt);
+  if(cp->flags & CPFLAG_TIMEOUT_OVERRIDES)
+    printf("Timeout: %d/18 seconds\n", cp->timeout);
+  if(cp->flags & CPFLAG_ENABLED)
   {
     printf("Enabled:");
-    if(cp.enabled & CPE_1)
+    if(cp->enabled & CPE_1)
       printf(" 1");
-    if(cp.enabled & CPE_2)
+    if(cp->enabled & CPE_2)
       printf(" 2");
-    if(cp.enabled & CPE_3)
+    if(cp->enabled & CPE_3)
       printf(" 3");
-    if(cp.enabled & CPE_4)
+    if(cp->enabled & CPE_4)
       printf(" 4");
-    if(cp.enabled & CPE_F)
+    if(cp->enabled & CPE_F)
       printf(" F");
-    if(cp.enabled & CPE_A)
+    if(cp->enabled & CPE_A)
       printf(" A");
       printf("\n");
   }
-  if(cp.flags & CPFLAG_INTERRUPT)
+  if(cp->flags & CPFLAG_INTERRUPT)
   {
     printf("Interrupt:");
-    if(cp.interrupt & CPI_SHIFT)
+    if(cp->interrupt & CPI_SHIFT)
       printf(" Shift");
-    if(cp.interrupt & CPI_KEY)
+    if(cp->interrupt & CPI_KEY)
       printf(" Key");
-    if(cp.interrupt & CPI_ALWAYS)
+    if(cp->interrupt & CPI_ALWAYS)
       printf(" Always");
     printf("\n");
   }
 }
 
-/* Copy parameters from src to dest.  Don't complain if quiet is
-   set.  Return true if copy failed. */
-static int
-copy_params(struct desc *dest, struct desc *src, int quiet)
+static void
+show_mbr(struct file_desc *dsc)
 {
   struct change_params cp;
-
-#if 0
-  fprintf(stderr, "Fetching parameters from %s\n", src->path);
-#endif
-
-  /* Do the copying via the version-neutral "change_params" structure */
-  if(fetch_params(&cp, src, quiet))
-    return 1;
-
-#if 0
-  fprintf(stderr, "Writing parameters to %s\n", dest->path);
-#endif
-  modify_params(dest, &cp);
-  return 0;
+  fetch_params(&cp, dsc, 0);
+  show_params(&cp);
 }
 
 static void
-copy_code(struct desc *dest, const struct desc *src)
+copy_code(struct file_desc *dest, const struct file_desc *src)
 {
   memcpy(dest->data, src->data, CODE_SIZE);
   /* Any errors should be reported against the source. */
@@ -335,7 +673,7 @@ copy_code(struct desc *dest, const struct desc *src)
 
 /* Copy the partition table */
 static void
-copy_table(struct desc *dest, const struct desc *src)
+copy_table(struct file_desc *dest, const struct file_desc *src)
 {
   memcpy(dest->data+PTN_TABLE_OFFSET, src->data+PTN_TABLE_OFFSET,
 	 PTN_TABLE_SIZE);
@@ -346,7 +684,7 @@ copy_table(struct desc *dest, const struct desc *src)
 typedef ssize_t io_func(int, void *, size_t);
 
 static size_t
-do_io(io_func *fn, const struct desc *dsc, int fd, size_t min_size)
+do_io(io_func *fn, const struct file_desc *dsc, int fd, size_t min_size)
 {
   int res;
   int done=0;
@@ -373,7 +711,7 @@ do_io(io_func *fn, const struct desc *dsc, int fd, size_t min_size)
 }
 
 static size_t
-read_from_fd(const struct desc *dsc, int fd, size_t min_size)
+read_from_fd(const struct file_desc *dsc, int fd, size_t min_size)
 {
   return do_io(read, dsc, fd, min_size);
 }
@@ -387,13 +725,13 @@ my_write(int fd, void *buf, size_t size)
 }
 
 static size_t
-write_to_fd(const struct desc *dsc, int fd, size_t min_size)
+write_to_fd(const struct file_desc *dsc, int fd, size_t min_size)
 {
   return do_io(my_write, dsc, fd, min_size);
 }
 
 static void
-read_file(struct desc *dsc)
+read_file(struct file_desc *dsc)
 {
   int fd=open(dsc->path, O_RDONLY|O_NOCTTY);
   if(fd==-1)
@@ -402,7 +740,7 @@ read_file(struct desc *dsc)
 	    prog_name, dsc->path, strerror(errno));
     exit(1);
   }
-  if(dsc->flags & DESC_FLAG_FREE_DATA)
+  if(dsc->flags & FILE_DESC_FLAG_FREE_DATA)
     free(dsc->data);
   dsc->data=malloc(512);
   if(dsc->data==NULL)
@@ -410,7 +748,7 @@ read_file(struct desc *dsc)
     fprintf(stderr, "%s: Malloc failed\n", prog_name);
     exit(1);
   }
-  dsc->flags|=DESC_FLAG_FREE_DATA;
+  dsc->flags|=FILE_DESC_FLAG_FREE_DATA;
   read_from_fd(dsc, fd, BUFFER_SIZE);
   if(close(fd)==-1)
   {
@@ -425,35 +763,30 @@ read_file(struct desc *dsc)
 static void
 install_mbr(struct data *dp)
 {
-  uchar *code, *params;
+  ui8_t *code, *params;
+  struct change_params new_params;
   int mbr_fd = -1;
   size_t mbr_size;
-  uchar mbr_in[BUFFER_SIZE];
-  uchar mbr_out[BUFFER_SIZE];
-  struct desc target_orig={mbr_in, dp->target.path, 0};
+  ui8_t mbr_in[BUFFER_SIZE];
+  ui8_t mbr_out[BUFFER_SIZE];
+  struct file_desc target_orig={mbr_in, dp->target.path, 0};
+
+  memset(mbr_in, 0, BUFFER_SIZE);
 
   dp->target.data=mbr_out;
 
-  /* Read the code file if specified.  Otherwise, use the
-     built-in version. */
-  if(dp->install.flags & DESC_FLAG_SPECIFIED)
-    read_file(&(dp->install));
-  else
-  {
-    dp->install.path="<internal>";
-    dp->install.data=MBR_START;
-    dp->install.flags=0;
-  }
-
-  /* Read the parameter file if specified. */
-  if(dp->params.flags & DESC_FLAG_SPECIFIED)
-    read_file(&(dp->params));
-  else
-  {
-    dp->params.path=dp->install.path;
-    dp->params.data=dp->install.data;
-    dp->params.flags=0;
-  }
+  /* We build up a change_params structure for all the parameters, so
+   * we can pick out things like the Y2K flag...
+   *
+   * Priorities:
+   *  1. Command line options.
+   *  2. Params file (if specified).
+   *  3. Original mbr unless -r specified.
+   *  4. Installed mbr (even if -k specified).
+   *
+   * We process these from 3 down to 1, and then apply the result to
+   * 4.
+   */
 
   /* Start with an empty record. */
   memset(mbr_out, 0, sizeof(mbr_out));
@@ -494,7 +827,67 @@ install_mbr(struct data *dp)
     exit(1);
   }
 
-  if(dp->table.flags & DESC_FLAG_SPECIFIED)
+  /* Try to extract the parameters unless we're resetting them.  If we
+     are resetting, or the extraction fails, reset them anyway. */
+  if ((dp->flags & DFLAG_RESET) != 0 ||
+      !fetch_params(&new_params, &target_orig, 1))
+  {
+    /* These aren't really needed. */
+    new_params.compat_version = 0;
+    new_params.version = 0;
+    /* Reset everything. */
+    new_params.flags = 0;
+    new_params.interrupt = 0;
+    new_params.interrupt_mask = ~0;
+    new_params.enabled = 0;
+    new_params.enabled_mask = ~0;
+  }
+
+  /* Read the parameter file if specified. */
+  if(dp->params.flags & FILE_DESC_FLAG_SPECIFIED)
+  {
+    struct change_params extra_params;
+    read_file(&(dp->params));
+    fetch_params(&extra_params, &(dp->params), 0);
+    merge_changes(&new_params, &extra_params);
+  }
+  else
+  {
+    dp->params.path=0;
+    dp->params.data=0;
+    dp->params.flags=0;
+  }
+
+  /* If a file to install is specified, ignore the Y2K setting from
+     the original target and the params file. */
+  if (dp->install.flags & FILE_DESC_FLAG_SPECIFIED)
+    new_params.flags &= ~(CPFLAG_Y2KBUG|CPFLAG_Y2KDATE);
+
+  /* Merge in the command line changes. */
+  merge_changes(&new_params, &(dp->changes));
+
+  /* Read the code file if specified.  Otherwise, use the
+     built-in version. */
+  if(dp->install.flags & FILE_DESC_FLAG_SPECIFIED)
+    read_file(&(dp->install));
+  else
+  {
+    /* Check if the Y2K flag is set.  */
+    if (new_params.flags & CPFLAG_Y2KBUG ?
+	new_params.y2k_fix : 0)
+    {
+      dp->install.path="<internal-y2k>";
+      dp->install.data=Y2K_START;
+    }
+    else
+    {
+      dp->install.path="<internal>";
+      dp->install.data=MBR_START;
+    }
+    dp->install.flags=0;
+  }
+
+  if(dp->table.flags & FILE_DESC_FLAG_SPECIFIED)
     read_file(&(dp->table));
   else
   {
@@ -511,36 +904,26 @@ install_mbr(struct data *dp)
     copy_code(&(dp->target), &(dp->install));
   } else {
     if(dp->flags & DFLAG_VERBOSE)
-      fprintf(stderr, "Copying code from %s\n", &target_orig);
+      fprintf(stderr, "Copying code from %s\n", target_orig.path);
     copy_code(&(dp->target), &target_orig);
   }
 
-  /* Install parameters from params if the user wants it, or 
-     we fail to copy parameters from the input. */
-  if((dp->params.flags & DESC_FLAG_SPECIFIED)!=0
-     || ((dp->flags & DFLAG_RESET)==0 &&
-	 copy_params(&(dp->target), &target_orig, 1)))
+  /* Modify the parameters according to the change_params structure. */
+  if(new_params.flags)
   {
     if(dp->flags & DFLAG_VERBOSE)
-      fprintf(stderr,"Resetting parameters from %s\n",
-	      dp->params.path);
-    copy_params(&(dp->target), &(dp->params), 0);
-  }
-  
-  if(dp->flags & DFLAG_VERBOSE)
-    fprintf(stderr,"Modifying parameters.\n");
+      fprintf(stderr,"Modifying parameters.\n");
 
-  /* Modify the parameters according to the change_params structure. */
-  if(dp->changes.flags)
-    modify_params(&(dp->target), &(dp->changes));
+    modify_params(&(dp->target), &new_params);
+  }
 
   /* Copy the partition table. */
   copy_table(&(dp->target), &(dp->table));
 
   if(dp->flags & DFLAG_LIST)
-  {
     show_mbr(&(dp->target));
-  }
+
+  /* Stamp the boot signature on to the result (just in case). */
   mbr_out[BUFFER_SIZE-2]=0x55;
   mbr_out[BUFFER_SIZE-1]=0xAA;
 
@@ -615,6 +998,9 @@ show_help(void)
 	 "Don't install anything.\n"
 	 "  -o <offset>, --offset <offset>    "
 	 "Install the MBR at byte offset <offset>.\n"
+	 "  -y[u|l|-], --y2kbug[=utc|=local|=off]\n"
+	 "                                    "
+	 "Fix a Y2K bug of some BIOS.\n"
 	 "  -P <path>, --parameters <path>    "
 	 "Get parameters from <path>.\n"
 	 "  -r, --reset                       "
@@ -665,7 +1051,7 @@ parse_enabled(struct change_params *params, const char *arg)
   while(*arg)
   {
     char c=*arg++;
-    uchar bit=0;
+    ui8_t bit=0;
     switch(c)
     {
     case '=':
@@ -713,7 +1099,7 @@ parse_interrupt(struct change_params *params, const char *arg)
   while(*arg)
   {
     char c=*arg++;
-    uchar bit=0;
+    ui8_t bit=0;
     switch(c)
     {
     case '=':
@@ -729,7 +1115,7 @@ parse_interrupt(struct change_params *params, const char *arg)
     case 'k': case 'K': bit=CPI_KEY; break;
     case 'a': case 'A': bit=CPI_ALWAYS;
       if(mode!='-') /* Don't enforce the timeout */
-	params->flags&=~CPFLAG_TIMEOUT;
+	params->flags&=~CPFLAG_TIMEOUT_OVERRIDES;
       break;
     case 'n': case 'N':
       params->interrupt_mask&=~CPI_MASK;
@@ -761,6 +1147,8 @@ parse_options(struct data *my_data, int argc, char *argv[])
 {
   int c;
   char to_do=0;
+  time_t atime;
+  struct tm btime;
   static struct option long_options[] =
   {
     /* {name, has_arg, flag, val} */
@@ -770,6 +1158,7 @@ parse_options(struct data *my_data, int argc, char *argv[])
     {"list",       0, 0, 'l'},
     {"no-act",     0, 0, 'n'},
     {"offset",     1, 0, 'o'},
+    {"y2kbug",     2, 0, 'y'},
     {"parameters", 1, 0, 'P'},
     {"reset",      0, 0, 'r'},
     {"table",      1, 0, 'T'},
@@ -794,10 +1183,11 @@ parse_options(struct data *my_data, int argc, char *argv[])
   while(1)
   {
     int long_index=0;
-    c=getopt_long(argc, argv, "fI:klno:P:rT:vVhd:e:i:p:t:",
+    c=getopt_long(argc, argv, "fI:klno:y::P:rT:vVhd:e:i:p:t:",
 		  long_options, &long_index);
     if(c==EOF)
       break;
+
     switch(c)
     {
     case 'f':
@@ -805,14 +1195,14 @@ parse_options(struct data *my_data, int argc, char *argv[])
       break;
 
     case 'I': /* code to install */
-      my_data->install.path=optarg;
-      my_data->install.flags=DESC_FLAG_SPECIFIED;
       to_do=1;
+      my_data->install.path=optarg;
+      my_data->install.flags=FILE_DESC_FLAG_SPECIFIED;
       break;
 
     case 'k': /* Keep the current code */
-      my_data->flags |= DFLAG_KEEP;
       to_do=1;
+      my_data->flags |= DFLAG_KEEP;
       break;
 
     case 'l': /* List mode */
@@ -827,21 +1217,97 @@ parse_options(struct data *my_data, int argc, char *argv[])
       my_data->offset=get_uint(optarg, "Offset");
       break;
 
-    case 'P': /* Parameters from file */
-      my_data->params.path=optarg;
-      my_data->params.flags=DESC_FLAG_SPECIFIED;
+    case 'y': /* y2kbug */
       to_do=1;
+      if (!optarg)
+      {
+	/* Try to determine the which timezone the RTC is set to. */
+	FILE *debdef=fopen("/etc/default/rcS", "r");
+	/* Default to UTC */
+	optarg="u";
+	if(debdef == NULL)
+	{
+	  if (errno!=ENOENT)
+	  {
+	    perror("Could not open /etc/default/rcS");
+	    exit(1);
+	  }
+	}
+	else
+	{
+	  char line[8];
+	  int c;
+	  while (fgets(line, 8, debdef))
+	  {
+	    if (!memcmp("UTC=no\n", line, 7)) {
+	      optarg="l";
+	      break;
+	    }
+	    if (strchr(line, '\n') == NULL)
+	    {
+	      do { c = getc(debdef); } while (c != EOF && c != '\n');
+	      if (c == EOF)
+		break;
+	    }
+	  }
+	  if (ferror (debdef) || fclose(debdef))
+	  {
+	    perror("Error reading /etc/default/rcS");
+	    exit(1);
+	  }
+	}
+      }
+      if (!strcmp(optarg, "-") ||
+	  !strcmp(optarg, "off"))
+      {
+	/* Request to turn Y2K bug off. */
+	my_data->changes.flags |= CPFLAG_Y2KBUG;
+	my_data->changes.y2k_fix = 0;
+      }
+      else
+      {
+	if (time(&atime) == (time_t)-1)
+	{
+	  fprintf(stderr, "%s: Error reading current time: %s\n",
+		  prog_name, strerror(errno));
+	  exit(1);
+	}
+	/* Valid arguments are "l", "u", "local" and "utc". */
+	if (optarg[0] == 'l'?optarg[1]!=0&&strcmp(optarg+1,"ocal"):
+	    optarg[0] == 'u'?optarg[1]!=0&&strcmp(optarg+1,"tc"):
+	    1)
+	{
+	  fprintf(stderr, "%s: Invalid timezone after Y2K option: %s\n",
+		  prog_name, optarg);
+	  exit(1);
+	}
+        if (optarg[0]=='l')
+          btime = *localtime(&atime);
+        else
+          btime = *gmtime(&atime);
+	/* We've got the date.  Store the changes. */
+	my_data->changes.flags |= CPFLAG_Y2KDATE | CPFLAG_Y2KBUG;
+	my_data->changes.y2k_fix = 1;
+	my_data->changes.y2k_month = btime.tm_mon+1;
+        my_data->changes.y2k_year = btime.tm_year+1900;
+      }
+      break;
+
+    case 'P': /* Parameters from file */
+      to_do=1;
+      my_data->params.path=optarg;
+      my_data->params.flags=FILE_DESC_FLAG_SPECIFIED;
       break;
 
     case 'r': /* Reset parameters */
-      my_data->flags |= DFLAG_RESET;
       to_do=1;
+      my_data->flags |= DFLAG_RESET;
       break;
 
     case 'T': /* Partition table from file */
-      my_data->table.path=optarg;
-      my_data->table.flags=DESC_FLAG_SPECIFIED;
       to_do=1;
+      my_data->table.path=optarg;
+      my_data->table.flags=FILE_DESC_FLAG_SPECIFIED;
       break;
 
     case 'v': /* verbose */
@@ -850,7 +1316,7 @@ parse_options(struct data *my_data, int argc, char *argv[])
       
     case 'V': /* version */
       printf("install-mbr %s\n", VERSION);
-      printf("Copyright (C) 1999 Neil Turton.\n");
+      printf(COPYRIGHT "\n");
       exit(0);
       break;
 
@@ -859,6 +1325,7 @@ parse_options(struct data *my_data, int argc, char *argv[])
       exit(0);
 
     case 'd': /* drive */
+      to_do=1;
       {
 	unsigned int drive=get_uint(optarg, "Drive");
 	if(drive>0xfe)
@@ -869,21 +1336,21 @@ parse_options(struct data *my_data, int argc, char *argv[])
 	};
 	my_data->changes.flags|=CPFLAG_DRIVE;
 	my_data->changes.drive=drive;
-	to_do=1;
       }
       break;
 
     case 'e':
-      parse_enabled(&(my_data->changes), optarg);
       to_do=1;
+      parse_enabled(&(my_data->changes), optarg);
       break;
 
     case 'i':
-      parse_interrupt(&(my_data->changes), optarg);
       to_do=1;
+      parse_interrupt(&(my_data->changes), optarg);
       break;
 
     case 'p':
+      to_do=1;
       if(optarg[0]==0 || optarg[1]!=0)
       {
 	fprintf(stderr, "%s: Invalid partition number: %d\n",
@@ -892,27 +1359,26 @@ parse_options(struct data *my_data, int argc, char *argv[])
       }
       my_data->changes.flags|=CPFLAG_DEFAULT;
       my_data->changes.deflt=*optarg;
-      to_do=1;
       break;
 
     case 't':
+      to_do=1;
       {
 	unsigned int timeout=get_uint(optarg, "Timeout");
-	if(timeout>0xfe)
+	if(timeout>0xfffe)
 	{
 	  fprintf(stderr, "%s: Invalid timeout: %d\n",
 		  prog_name, timeout);
 	  exit(1);
 	}
-	my_data->changes.flags|=CPFLAG_TIMEOUT_VALID|CPFLAG_TIMEOUT;
+	my_data->changes.flags|=CPFLAG_TIMEOUT_VALID|CPFLAG_TIMEOUT_OVERRIDES;
 	my_data->changes.timeout=timeout;
-	to_do=1;
       }
       break;
 
     default:
       fprintf(stderr, "%s: Bad return from getopt_long: %d(%c)\n",
-	      prog_name, c);
+	      prog_name, errno, c);
       /* Fall-through */
     case '?':
       suggest_help();
@@ -926,16 +1392,17 @@ parse_options(struct data *my_data, int argc, char *argv[])
     exit(1);
   }
   my_data->target.path=argv[optind];
-  my_data->target.flags=DESC_FLAG_SPECIFIED;
+  my_data->target.flags=FILE_DESC_FLAG_SPECIFIED;
 
   /* Don't do anything if all we're asked to do is list. */
   if((my_data->flags & DFLAG_LIST) && !to_do)
-    my_data->flags |= DFLAG_NO_ACT;
+    my_data->flags |= DFLAG_NO_ACT | DFLAG_KEEP;
 }
 
 int
 main(int argc, char *argv[])
 {
+  /* FIXME: Use named fields */
   struct data my_data={
     /* install */
     {NULL, NULL, 0},
@@ -950,10 +1417,11 @@ main(int argc, char *argv[])
     /* flags */
     0,
     /* change_params structure */
-    {0, 0, 0, 0, 0, 0xff, 0, 0xff }};
+    {0, 0, 0, 0, 0, 0, 0, 0xff, 0, 0xff, 0 }};
 
   assert(MBR_SIZE==BUFFER_SIZE);
-
+  assert(Y2K_SIZE==BUFFER_SIZE);
+  
   parse_options(&my_data, argc, argv);
   install_mbr(&my_data);
   return 0;
